import warnings
from typing import Dict, List

import numpy as np
from scipy.optimize import minimize

from ...C import (
    CURRENT_SIMULATION,
    DATAPOINTS,
    EXPDATA_MASK,
    MAX_DATAPOINT,
    MIN_DATAPOINT,
    MIN_DIFF_FACTOR,
    MIN_SIM_RANGE,
    N_SPLINE_PARS,
    NOISE_PARAMETERS,
    NUM_DATAPOINTS,
    SCIPY_FUN,
    SCIPY_SUCCESS,
    SCIPY_X,
    InnerParameterType,
)
from ..solver import InnerSolver
from .parameter import SplineInnerParameter
from .problem import SplineInnerProblem

try:
    from amici.parameter_mapping import ParameterMapping
except ImportError:
    pass


class SplineInnerSolver(InnerSolver):
    """Solver of the inner subproblem of spline approximation for nonlinear-monotone data.

    Options
    -------
    min_diff_factor:
        Determines the minimum difference between two consecutive spline
        as ``min_diff_factor * (measurement_range) / n_spline_pars``.
        Default is 1/2.
    """

    def __init__(self, options: Dict = None):
        self.options = {
            **self.get_default_options(),
            **(options or {}),
        }
        self.validate_options()

    def validate_options(self):
        """Validate the current options dictionary."""
        if type(self.options[MIN_DIFF_FACTOR]) is not float:
            raise TypeError(f"{MIN_DIFF_FACTOR} must be of type float.")
        elif self.options[MIN_DIFF_FACTOR] < 0:
            raise ValueError(f"{MIN_DIFF_FACTOR} must be greater than zero.")
        for key in self.options:
            if key not in self.get_default_options():
                raise ValueError(f"Unknown SplineInnerSolver option {key}.")

    def solve(
        self,
        problem: SplineInnerProblem,
        sim: List[np.ndarray],
        sigma: List[np.ndarray],
    ) -> list:
        """Get results for every group (inner optimization problem).

        Parameters
        ----------
        problem:
            InnerProblem from pyPESTO hierarchical.
        sim:
            Simulations from AMICI.
        sigma:
            List of sigmas from AMICI.

        Returns
        -------
        List of optimization results of the inner subproblem.
        """
        inner_optimization_results = []
        for group in problem.get_groups_for_xs(InnerParameterType.SPLINE):
            group_dict = problem.groups[group]
            group_dict[NOISE_PARAMETERS] = extract_expdata_using_mask(
                expdata=sigma, mask=group_dict[EXPDATA_MASK]
            )
            group_dict[CURRENT_SIMULATION] = extract_expdata_using_mask(
                expdata=sim, mask=group_dict[EXPDATA_MASK]
            )

            inner_optimization_results_per_group = self._optimize_spline(
                inner_parameters=problem.get_free_xs_for_group(group),
                group_dict=group_dict,
            )
            inner_optimization_results.append(
                inner_optimization_results_per_group
            )
            save_inner_parameters_to_inner_problem(
                inner_parameters=problem.get_xs_for_group(group),
                s=inner_optimization_results_per_group[SCIPY_X],
            )
        return inner_optimization_results

    @staticmethod
    def calculate_obj_function(x_inner_opt: list):
        """Calculate the inner objective function value.

        Calculates the inner objective function value from a list of inner
        optimization results returned from `_optimize_spline`.

        Parameters
        ----------
        x_inner_opt:
            List of optimization results

        Returns
        -------
        Inner objective function value.
        """
        if False in (
            x_inner_opt[idx][SCIPY_SUCCESS] for idx in range(len(x_inner_opt))
        ):
            obj = np.inf
            warnings.warn("Inner optimization failed.")
        else:
            obj = np.sum(
                [
                    x_inner_opt[idx][SCIPY_FUN]
                    for idx in range(len(x_inner_opt))
                ]
            )
        return obj

    def calculate_gradients(
        self,
        problem: SplineInnerProblem,
        x_inner_opt: List[Dict],
        sim: List[np.ndarray],
        sigma: List[np.ndarray],
        sy: List[np.ndarray],
        parameter_mapping: ParameterMapping,
        par_opt_ids: List,
        par_sim_ids: List,
        snllh: Dict,
    ):
        """Calculate gradients of the inner objective function.

        Calculates gradients of the objective function with respect to outer
        (dynamical) parameters.

        Parameters
        ----------
        problem:
            Optimal scaling inner problem.
        x_inner_opt:
            List of optimization results of the inner subproblem.
        sim:
            Model simulations.
        sigma:
            Model noise parameters.
        sy:
            Model sensitivities.
        parameter_mapping:
            Mapping of optimization to simulation parameters.
        par_opt_ids:
            Ids of outer otimization parameters.
        par_sim_ids:
            Ids of outer simulation parameters, includes fixed parameters.
        snllh:
            Empty dictionary with optimization parameters as keys.

        Returns
        -------
        Filled in snllh dictionary with objective function gradients.
        """
        already_calculated = set()

        for condition_map_sim_var in [
            cond_par_map.map_sim_var for cond_par_map in parameter_mapping
        ]:
            for par_sim, par_opt in condition_map_sim_var.items():
                if (
                    not isinstance(par_opt, str)
                    or par_opt in already_calculated
                ):
                    continue
                # Current fix for scaling/offset parameters in models.
                elif par_sim.startswith('observableParameter'):
                    continue
                else:
                    already_calculated.add(par_opt)
                par_sim_idx = par_sim_ids.index(par_sim)
                par_opt_idx = par_opt_ids.index(par_opt)
                grad = 0.0
                sy_for_outer_parameter = [
                    sy_cond[:, par_sim_idx, :] for sy_cond in sy
                ]

                for group_idx, group in enumerate(
                    problem.get_groups_for_xs(InnerParameterType.SPLINE)
                ):
                    # Get the reformulated spline parameters
                    s = np.asarray(x_inner_opt[group_idx][SCIPY_X])
                    group_dict = problem.groups[group]

                    measurements = group_dict[DATAPOINTS]
                    sigma = group_dict[NOISE_PARAMETERS]
                    sim_all = group_dict[CURRENT_SIMULATION]
                    N = group_dict[N_SPLINE_PARS]
                    K = group_dict[NUM_DATAPOINTS]

                    sy_all = extract_expdata_using_mask(
                        expdata=sy_for_outer_parameter,
                        mask=group_dict[EXPDATA_MASK],
                    )

                    delta_c, c, n = self._rescale_spline_bases(
                        sim_all=sim_all, N=N, K=K
                    )
                    delta_c_dot, c_dot = calculate_spline_bases_gradient(
                        sim_all=sim_all, sy_all=sy_all, N=N
                    )
                    C = np.diag(-np.ones(N))

                    # For the reformulated problem, mu can be calculated
                    # as the inner gradient at the optimal point s.
                    mu = calculate_inner_gradient(
                        s=s,
                        sim_all=sim_all,
                        measurements=measurements,
                        sigma=sigma,
                        N=N,
                        delta_c=delta_c,
                        c=c,
                        n=n,
                    )
                    min_meas = group_dict[MIN_DATAPOINT]
                    max_meas = group_dict[MAX_DATAPOINT]
                    min_diff = self._get_minimal_difference(
                        measurement_range=max_meas - min_meas,
                        N=N,
                        min_diff_factor=self.options[MIN_DIFF_FACTOR],
                    )

                    # Calculate df_ds term only if mu is not all 0
                    if np.any(mu):
                        s_dot = calculate_ds_dtheta(
                            sim_all=sim_all,
                            sy_all=sy_all,
                            measurements=measurements,
                            sigma=sigma,
                            s=s,
                            C=C,
                            mu=mu,
                            N=N,
                            delta_c=delta_c,
                            delta_c_dot=delta_c_dot,
                            c=c,
                            c_dot=c_dot,
                            n=n,
                            min_diff=min_diff,
                        )
                        df_ds = mu
                        grad += df_ds.dot(s_dot)

                    # Let's calculate the df_dyk term now:
                    df_dyk = calculate_df_dyk(
                        sim_all=sim_all,
                        sy_all=sy_all,
                        measurements=measurements,
                        sigma=sigma,
                        s=s,
                        N=N,
                        delta_c=delta_c,
                        delta_c_dot=delta_c_dot,
                        c=c,
                        c_dot=c_dot,
                        n=n,
                    )

                    grad += df_dyk
                snllh[par_opt_idx] = grad

        return snllh

    @staticmethod
    def get_default_options() -> Dict:
        """Return default options for solving the inner problem."""
        options = {
            MIN_DIFF_FACTOR: 1 / 2,
        }
        return options

    def _optimize_spline(
        self,
        inner_parameters: List[SplineInnerParameter],
        group_dict: Dict,
    ):
        """Run optimization for the inner problem.

        Parameters
        ----------
        inner_parameters:
            The spline inner parameters.
        group_dict:
            The group dictionary.
        """
        group_measurements = group_dict[DATAPOINTS]
        group_noise_parameters = group_dict[NOISE_PARAMETERS]
        current_group_simulation = group_dict[CURRENT_SIMULATION]
        n_datapoints = group_dict[NUM_DATAPOINTS]
        n_spline_pars = group_dict[N_SPLINE_PARS]

        (
            distance_between_bases,
            spline_bases,
            intervals_per_sim,
        ) = self._rescale_spline_bases(
            sim_all=current_group_simulation, N=n_spline_pars, K=n_datapoints
        )

        min_meas = group_dict[MIN_DATAPOINT]
        max_meas = group_dict[MAX_DATAPOINT]
        min_diff = self._get_minimal_difference(
            measurement_range=max_meas - min_meas,
            N=n_spline_pars,
            min_diff_factor=self.options[MIN_DIFF_FACTOR],
        )

        inner_options = self._get_inner_optimization_options(
            inner_parameters=inner_parameters,
            N=n_spline_pars,
            min_meas=min_meas,
            max_meas=max_meas,
            min_diff=min_diff,
        )

        def objective_function_wrapper(x):
            return calculate_objective_function(
                s=x,
                sim_all=current_group_simulation,
                measurements=group_measurements,
                sigma=group_noise_parameters,
                N=n_spline_pars,
                delta_c=distance_between_bases,
                c=spline_bases,
                n=intervals_per_sim,
            )

        def inner_gradient_wrapper(x):
            return calculate_inner_gradient(
                s=x,
                sim_all=current_group_simulation,
                measurements=group_measurements,
                sigma=group_noise_parameters,
                N=n_spline_pars,
                delta_c=distance_between_bases,
                c=spline_bases,
                n=intervals_per_sim,
            )

        results = minimize(
            objective_function_wrapper,
            jac=inner_gradient_wrapper,
            **inner_options,
        )

        return results

    def _rescale_spline_bases(self, sim_all: np.ndarray, N: int, K: int):
        """Rescale the spline bases.

        Before the optimization of the spline parameters, we have to fix the
        spline bases to some values. We choose to scale them to the current
        simulation. In case of simulations that are very close to each other,
        we choose to scale closely around the average value of the simulations,
        to avoid numerical problems (as we often divide by delta_c).

        Parameters
        ----------
        sim_all:
            The current simulation.
        N:
            The number of spline parameters.
        K:
            The number of simulations.

        Returns
        -------
        distance_between_bases:
            The distance between the spline bases.
        spline_bases:
            The rescaled spline bases.
        intervals_per_sim:
            List of indices of intervals each simulation belongs to.
        """
        min_idx = np.argmin(sim_all)
        max_idx = np.argmax(sim_all)

        min_all = sim_all[min_idx]
        max_all = sim_all[max_idx]

        n = np.ones(K)

        # In case the simulation are very close to each other
        # or even collapse into a single point.
        if max_all - min_all < MIN_SIM_RANGE:
            average_value = (max_all + min_all) / 2
            delta_c = MIN_SIM_RANGE / (N - 1)
            if average_value < (MIN_SIM_RANGE / 2):
                c = np.linspace(0, MIN_SIM_RANGE, N)
            else:
                c = np.linspace(
                    average_value - (MIN_SIM_RANGE / 2),
                    average_value + (MIN_SIM_RANGE / 2),
                    N,
                )
            # Set the n(k) values for the simulations
            for i in range(len(sim_all)):
                n[i] = np.ceil((sim_all[i] - c[0]) / delta_c) + 1
                if n[i] > N:
                    n[i] = N
                    warnings.warn(
                        "Interval for a simulation has been set to a larger value than the number of spline parameters."
                    )
        # In case the simulations are sufficiently apart:
        else:
            delta_c = (max_all - min_all) / (N - 1)
            c = np.linspace(min_all, max_all, N)
            for i in range(len(sim_all)):
                if i == max_idx:
                    n[i] = N
                elif i == min_idx:
                    n[i] = 1
                else:
                    n[i] = np.ceil((sim_all[i] - c[0]) / delta_c) + 1
                if n[i] > N:
                    n[i] = N

        n = n.astype(int)
        return delta_c, c, n

    def _get_minimal_difference(
        self,
        measurement_range: float,
        N: int,
        min_diff_factor: float,
    ):
        """Return minimal parameter difference for spline parameters."""
        return min_diff_factor * measurement_range / N

    def _get_inner_optimization_options(
        self,
        inner_parameters: List[SplineInnerParameter],
        N: int,
        min_meas: float,
        max_meas: float,
        min_diff: float,
    ) -> Dict:
        """Return default options for scipy optimizer.

        Returns inner subproblem optimization options including startpoint
        and optimization bounds or constraints, dependent on solver method.

        Parameters
        ----------
        inner_parameters:
            Inner parameters of the spline group.
        N:
            Number of spline parameters.
        min_meas:
            Minimal measurement value.
        max_meas:
            Maximal measurement value.
        min_diff:
            Minimal difference between spline parameters.
        """
        range_all = max_meas - min_meas

        constraint_min_diff = np.full(N, min_diff)
        constraint_min_diff[0] = 0

        last_opt_values = np.asarray([x.value for x in inner_parameters])

        if (last_opt_values > 0).any():
            x0 = last_opt_values
        # In case this is the first inner optimization, initialize the
        # spline parameters to a linear function with a symmetric 60%
        # larger range than the measurement range.
        else:
            x0 = np.full(
                N,
                (
                    max_meas
                    + 0.3 * range_all
                    - np.max([min_meas - 0.3 * range_all, 0])
                )
                / (N - 1),
            )
            x0[0] = np.max([min_meas - 0.3 * range_all, 0])

        from scipy.optimize import Bounds

        inner_options = {
            "x0": x0,
            "method": "L-BFGS-B",
            "options": {"ftol": 1e-10, "disp": None},
            "bounds": Bounds(lb=constraint_min_diff),
        }

        return inner_options


def calculate_objective_function(
    s: np.ndarray,
    sim_all: np.ndarray,
    measurements: np.ndarray,
    sigma: np.ndarray,
    N: int,
    delta_c: float,
    c: np.ndarray,
    n: np.ndarray,
):
    """Objective function for reformulated inner spline problem."""
    obj = 0

    for y_k, z_k, sigma_k, n_k in zip(sim_all, measurements, sigma, n):
        i = n_k - 1
        sum_s = 0
        sum_s = np.sum(s[:i])
        if i == 0:
            obj += (1 / sigma_k**2) * (z_k - s[i]) ** 2
        elif i == N:
            obj += (1 / sigma_k**2) * (z_k - sum_s) ** 2
        else:
            obj += (1 / sigma_k**2) * (
                z_k - (y_k - c[i - 1]) * s[i] / delta_c - sum_s
            ) ** 2
    obj = obj / 2
    return obj


def get_spline_mapped_simulations(
    s: np.ndarray,
    sim_all: np.ndarray,
    N: int,
    delta_c: float,
    c: np.ndarray,
    n: np.ndarray,
):
    """Return model simulations mapped using the approximation spline."""
    mapped_simulations = np.zeros(len(sim_all))
    xi = np.zeros(len(s))
    for i in range(len(s)):
        xi[i] = np.sum(s[: i + 1])

    for y_k, n_k, index in zip(sim_all, n, range(len(sim_all))):
        interval_index = n_k - 1
        if interval_index == 0 or interval_index == N:
            mapped_simulations[index] = xi[interval_index]
        else:
            mapped_simulations[index] = (y_k - c[interval_index - 1]) * (
                xi[interval_index] - xi[interval_index - 1]
            ) / delta_c + xi[interval_index - 1]

    return mapped_simulations


def calculate_inner_gradient(
    s: np.ndarray,
    sim_all: np.ndarray,
    measurements: np.ndarray,
    sigma: np.ndarray,
    N: int,
    delta_c: float,
    c: np.ndarray,
    n: np.ndarray,
):
    """Gradient of the objective function for the reformulated inner spline problem."""

    gradient = np.zeros(N)

    for y_k, z_k, sigma_k, n_k in zip(sim_all, measurements, sigma, n):
        weight_k = 1 / sigma_k**2
        sum_s = 0
        i = n_k - 1  # just the iterator to go over the Jacobian array
        sum_s = np.sum(s[:i])
        if i == 0:
            gradient[i] += weight_k * (s[i] - z_k)
        elif i == N:
            for j in range(i):
                gradient[j] += weight_k * (sum_s - z_k)
        else:
            gradient[i] += (
                weight_k
                * ((y_k - c[i - 1]) * s[i] / delta_c + sum_s - z_k)
                * (y_k - c[i - 1])
                / delta_c
            )
            gradient[:i] += np.full(
                i,
                weight_k * ((y_k - c[i - 1]) * s[i] / delta_c + sum_s - z_k),
            )
    return gradient


def calculate_inner_hessian(
    s: np.ndarray,
    sim_all: np.ndarray,
    sigma: np.ndarray,
    N: int,
    delta_c: float,
    c: np.ndarray,
    n: np.ndarray,
):
    """Calculate the hessian of the objective function for the reformulated inner problem."""

    hessian = np.zeros((N, N))

    for y_k, sigma_k, n_k in zip(sim_all, sigma, n):
        sum_s = 0
        i = n_k - 1  # just the iterator to go over the Hessian matrix
        for j in range(i):
            sum_s += s[j]

        hessian[i][i] += (1 / sigma_k**2) * ((y_k - c[i - 1]) / delta_c) ** 2
        for j in range(i):
            hessian[i][j] += (1 / sigma_k**2) * ((y_k - c[i - 1]) / delta_c)
            hessian[j][i] += (1 / sigma_k**2) * ((y_k - c[i - 1]) / delta_c)
            for h in range(i):
                hessian[j][h] += 1 / sigma_k**2

    return hessian


def calculate_ds_dtheta(
    sim_all: np.ndarray,
    sy_all: np.ndarray,
    measurements: np.ndarray,
    sigma: np.ndarray,
    s: np.ndarray,
    C: np.ndarray,
    mu: np.ndarray,
    N: int,
    delta_c: float,
    delta_c_dot: float,
    c: np.ndarray,
    c_dot: np.ndarray,
    n: np.ndarray,
    min_diff: float,
):
    """Calculate derivatives of reformulated spline parameters with respect to outer parameter.

    Calculates the derivative of reformulated spline parameters s with respect to the
    dynamical parameter theta. Firstly, we calculate the derivative of the
    first two equations of the necessary optimality conditions of the
    optimization problem with inequality constraints. Then we solve the linear
    system to obtain the derivatives.
    """

    Jacobian_derivative = np.zeros((N, N))
    rhs = np.zeros(2 * N)

    for y_k, z_k, y_dot_k, sigma_k, n_k in zip(
        sim_all, measurements, sy_all, sigma, n
    ):
        i = n_k - 1  # just the iterator to go over the Jacobian matrix
        weight_k = 1 / sigma_k**2
        sum_s = 0
        sum_s = np.sum(s[:i])

        # calculate the Jacobian derivative:
        if i == 0:
            Jacobian_derivative[i][i] += weight_k
        elif i == N:
            Jacobian_derivative = Jacobian_derivative + np.full(
                (N, N), weight_k
            )

        else:
            Jacobian_derivative[i][i] += (
                weight_k * (y_k - c[i - 1]) ** 2 / delta_c**2
            )
            rhs[i] += (
                weight_k
                * (2 * (y_k - c[i - 1]) / delta_c * s[i] + sum_s - z_k)
                * (
                    (y_dot_k - c_dot[i - 1]) * delta_c
                    - (y_k - c[i - 1]) * delta_c_dot
                )
                / delta_c**2
            )
            if i > 0:
                Jacobian_derivative[i, :i] += np.full(
                    i, weight_k * (y_k - c[i - 1]) / delta_c
                )
                Jacobian_derivative[:i, i] += np.full(
                    i, weight_k * (y_k - c[i - 1]) / delta_c
                )
                rhs[:i] += np.full(
                    i,
                    weight_k
                    * (
                        (y_dot_k - c_dot[i - 1]) * delta_c
                        - (y_k - c[i - 1]) * delta_c_dot
                    )
                    * s[i]
                    / delta_c**2,
                )
                Jacobian_derivative[:i, :i] += np.full((i, i), weight_k)

    from scipy import linalg

    constraint_min_diff = np.diag(np.full(N, min_diff))
    constraint_min_diff[0][0] = 0
    lhs = np.block(
        [
            [Jacobian_derivative, C],
            [-np.diag(mu), constraint_min_diff - np.diag(s)],
        ]
    )
    ds_dtheta = linalg.lstsq(lhs, rhs, lapack_driver="gelsy")

    return ds_dtheta[0][:N]


def calculate_df_dyk(
    sim_all: np.ndarray,
    sy_all: np.ndarray,
    measurements: np.ndarray,
    sigma: np.ndarray,
    s: np.ndarray,
    N: int,
    delta_c: float,
    delta_c_dot: float,
    c: np.ndarray,
    c_dot: np.ndarray,
    n: np.ndarray,
):
    """Calculate the derivative of the objective function for one group with respect to the simulations."""
    df_dyk = 0

    for y_k, z_k, y_dot_k, sigma_k, n_k in zip(
        sim_all, measurements, sy_all, sigma, n
    ):
        i = n_k - 1
        sum_s = np.sum(s[:i])
        if i > 0 and i < N:
            df_dyk += (
                (1 / sigma_k**2)
                * ((y_k - c[i - 1]) * s[i] / delta_c + sum_s - z_k)
                * s[i]
                * (
                    (y_dot_k - c_dot[i - 1]) * delta_c
                    - (y_k - c[i - 1]) * delta_c_dot
                )
                / delta_c**2
            )
    return df_dyk


def calculate_spline_bases_gradient(
    sim_all: np.ndarray, sy_all: np.ndarray, N: int
):
    """Calculate gradient of the rescaled spline bases."""

    min_idx = np.argmin(sim_all)
    max_idx = np.argmax(sim_all)

    if sim_all[max_idx] - sim_all[min_idx] < MIN_SIM_RANGE:
        delta_c_dot = 0
        c_dot = np.full(N, (sy_all[max_idx] - sy_all[min_idx]) / 2)
    else:
        delta_c_dot = (sy_all[max_idx] - sy_all[min_idx]) / (N - 1)
        c_dot = np.linspace(sy_all[min_idx], sy_all[max_idx], N)

    return delta_c_dot, c_dot


def extract_expdata_using_mask(
    expdata: List[np.ndarray], mask: List[np.ndarray]
):
    """Extract data from expdata list of arrays for the given mask."""
    return np.concatenate(
        [
            expdata[condition_index][mask[condition_index]]
            for condition_index in range(len(mask))
        ]
    )


def save_inner_parameters_to_inner_problem(
    inner_parameters,
    s,
) -> None:
    """Save inner parameter values to the inner subproblem.

    Calculates the non-reformulated inner spline parameters from
    the reformulated inner spline parameters and saves them to
    the inner subproblem.

    Parameters
    ----------
    inner_parameters : list
        List of inner parameters.
    s : np.ndarray
        Reformulated inner spline parameters.
    """
    xi = np.zeros(len(s))
    for i in range(len(s)):
        xi[i] = np.sum(s[: i + 1])

    for idx in range(len(inner_parameters)):
        inner_parameters[idx].value = xi[idx]


def get_monotonicity_measure(measurement, simulation):
    """Get monotonicity measure by calculating inversions.

    Calculate the number of inversions in the simulation data
    with respect to the measurement data.

    Parameters
    ----------
    measurement : np.ndarray
        Measurement data.
    simulation : np.ndarray
        Simulation data.

    Returns
    -------
    inversions : int
        Number of inversions.
    """
    if len(measurement) != len(simulation):
        raise ValueError(
            "Measurement and simulation data must have the same length."
        )

    ordered_simulation = [
        x
        for _, x in sorted(
            zip(measurement, simulation), key=lambda pair: pair[0]
        )
    ]
    ordered_measurement = sorted(simulation)

    inversions = 0
    for i in range(len(ordered_simulation)):
        for j in range(i + 1, len(ordered_simulation)):
            if ordered_simulation[i] > ordered_simulation[j]:
                inversions += 1
            elif (
                ordered_simulation[i] == ordered_simulation[j]
                and ordered_measurement[i] != ordered_measurement[j]
            ):
                inversions += 1

    return inversions



import warnings
from typing import Dict, List, Optional

import matplotlib.pyplot as plt
import numpy as np

try:
    import amici
    from petab.C import OBSERVABLE_ID

    from ..hierarchical.optimal_scaling.calculator import (
        OptimalScalingAmiciCalculator,
    )
    from ..hierarchical.optimal_scaling.parameter import (
        OptimalScalingParameter,
    )
    from ..hierarchical.optimal_scaling.problem import OptimalScalingProblem
    from ..hierarchical.optimal_scaling.solver import (
        OptimalScalingInnerSolver,
        compute_interval_constraints,
        get_bounds_for_category,
        undo_inner_parameter_reparameterization,
    )
except ImportError:
    pass


from ..C import (
    AMICI_SIGMAY,
    AMICI_T,
    AMICI_Y,
    CENSORED,
    MEASUREMENT_TYPE,
    ORDINAL,
    QUANTITATIVE_DATA,
    QUANTITATIVE_IXS,
    REPARAMETERIZED,
    SCIPY_X,
)
from ..result import Result


def plot_categories_from_pypesto_result(
    pypesto_result: Result,
    start_index=0,
    axes: Optional[plt.Axes] = None,
    **kwargs,
):
    """Plot the inner solutions from a pypesto result.

    Parameters
    ----------
    pypesto_result:
        The pypesto result.
    start_index:
        The index of the pypesto_result.optimize_result.list to plot.
    axes:
        The optional axes to plot on.
    kwargs:
        Additional arguments to pass to the figure.

    Returns
    -------
    fig:
        The figure.
    axes:
        The axes.
    """

    # Get the parameters from the pypesto result for the start_index.
    x_dct = dict(
        zip(
            pypesto_result.problem.objective.x_ids,
            pypesto_result.optimize_result.list[start_index]['x'],
        )
    )

    # Get the needed objects from the pypesto problem.
    edatas = pypesto_result.problem.objective.edatas
    parameter_mapping = pypesto_result.problem.objective.parameter_mapping
    amici_model = pypesto_result.problem.objective.amici_model
    amici_solver = pypesto_result.problem.objective.amici_solver
    petab_problem = (
        pypesto_result.problem.objective.amici_object_builder.petab_problem
    )
    n_threads = pypesto_result.problem.objective.n_threads

    # Fill in the parameters.
    amici.parameter_mapping.fill_in_parameters(
        edatas=edatas,
        problem_parameters=x_dct,
        scaled_parameters=True,
        parameter_mapping=parameter_mapping,
        amici_model=amici_model,
    )

    # Simulate the model with the parameters from the pypesto result.
    inner_rdatas = amici.runAmiciSimulations(
        amici_model,
        amici_solver,
        edatas,
        num_threads=min(n_threads, len(edatas)),
    )

    # If any amici simulation failed, raise warning and return None.
    if any(rdata.status != amici.AMICI_SUCCESS for rdata in inner_rdatas):
        warnings.warn(
            'Warning: Some AMICI simulations failed. Cannot plot inner solutions.'
        )
        return None

    # Get simulation and sigma.
    sim = [rdata[AMICI_Y] for rdata in inner_rdatas]
    sigma = [rdata[AMICI_SIGMAY] for rdata in inner_rdatas]
    timepoints = [rdata[AMICI_T] for rdata in inner_rdatas]
    observable_ids = amici_model.getObservableIds()
    condition_ids = [edata.id for edata in edatas]
    petab_condition_ordering = list(petab_problem.condition_df.index)

    # Get the observable ordering from the measurement_df.
    measurement_df_observable_ordering = list(
        petab_problem.measurement_df[OBSERVABLE_ID].unique()
    )

    optimal_scaling_calculator = None
    for (
        calculator
    ) in pypesto_result.problem.objective.calculator.inner_calculators:
        if isinstance(calculator, OptimalScalingAmiciCalculator):
            optimal_scaling_calculator = calculator
            break

    # Get the inner solver and problem.
    inner_solver = optimal_scaling_calculator.inner_solver
    inner_problem = optimal_scaling_calculator.inner_problem

    inner_results = inner_solver.solve(inner_problem, sim, sigma)

    return plot_categories_from_inner_result(
        inner_problem,
        inner_solver,
        inner_results,
        sim,
        timepoints,
        observable_ids,
        condition_ids,
        petab_condition_ordering,
        measurement_df_observable_ordering,
        axes,
        **kwargs,
    )


def plot_categories_from_inner_result(
    inner_problem: 'OptimalScalingProblem',
    inner_solver: 'OptimalScalingInnerSolver',
    results: List[Dict],
    simulation: List[np.ndarray],
    timepoints: List[np.ndarray],
    observable_ids: List[str] = None,
    condition_ids: List[str] = None,
    petab_condition_ordering: List[str] = None,
    measurement_df_observable_ordering: List[str] = None,
    axes: Optional[plt.Axes] = None,
    **kwargs,
):
    """Plot the inner solutions.

    Parameters
    ----------
    inner_problem:
        The inner problem.
    inner_solver:
        The inner solver.
    results:
        The results from the inner solver.
    simulation:
        The model simulation.
    timepoints:
        The timepoints of the simulation.
    kwargs:
        Additional arguments to pass to the figure.
    axes:
        The optional axes to plot on.

    Returns
    -------
    fig:
        The figure.
    axes:
        The axes.
    """

    if len(results) != len(inner_problem.groups):
        raise ValueError(
            "Number of results must be equal to number of groups of the inner subproblem."
        )

    # Get the number of groups
    n_groups = len(inner_problem.groups)
    options = inner_solver.options

    use_given_axes = axes is not None

    # If there are no axes, make a figure with multiple plots
    if axes is None:
        axes = _get_default_axes(n_groups, **kwargs)

    # for each result and group, plot the inner solution
    for result, group in zip(results, inner_problem.groups):
        if observable_ids is not None and use_given_axes:
            observable_id = observable_ids[group - 1]
            meas_obs_idx = measurement_df_observable_ordering.index(
                observable_id
            )

            # Get the ax for the current observable.
            ax = axes['plot' + str(meas_obs_idx + 1)]
        else:
            ax = axes[list(inner_problem.groups.keys()).index(group)]

        # For each group get the inner parameters and simulation
        xs = inner_problem.get_cat_ub_parameters_for_group(group)

        interval_range, interval_gap = compute_interval_constraints(
            xs, simulation, options
        )
        observable_index = group - 1
        measurement_type = inner_problem.groups[group][MEASUREMENT_TYPE]
        # Get surrogate datapoints and category bounds
        (
            simulation_all,
            surrogate_all,
            timepoints_all,
            upper_bounds_all,
            lower_bounds_all,
        ) = _get_data_for_plotting(
            xs,
            result[SCIPY_X],
            simulation,
            timepoints,
            interval_range,
            interval_gap,
            options,
            measurement_type,
        )

        # Get the number of distinct timepoints in timepoints_all
        # where timepoints_all is a list of numpy arrays of timepoints
        n_distinct_timepoints = len(np.unique(np.concatenate(timepoints_all)))

        # If there is only one distinct timepoint, plot with respect to conditions
        if n_distinct_timepoints == 1 and not use_given_axes:
            _plot_observable_fit_across_conditions(
                ax,
                inner_problem,
                observable_index,
                group,
                condition_ids,
                simulation,
                simulation_all,
                surrogate_all,
                upper_bounds_all,
                lower_bounds_all,
                measurement_type,
                petab_condition_ordering,
                use_given_axes,
            )

        # Plotting across timepoints
        elif n_distinct_timepoints > 1:
            n_conditions = len(simulation_all)

            # If there is only one condition, we don't need
            # separate colors for the different conditions
            if n_conditions == 1:
                _plot_observable_fit_for_one_condition(
                    ax,
                    observable_index,
                    group,
                    inner_problem,
                    timepoints,
                    timepoints_all,
                    simulation,
                    simulation_all,
                    surrogate_all,
                    lower_bounds_all,
                    upper_bounds_all,
                    measurement_type,
                    use_given_axes,
                )

            # If there are multiple conditions, we need
            # separate colors for the different conditions
            elif n_conditions > 1:
                _plot_observable_fit_for_multiple_conditions(
                    ax,
                    observable_index,
                    group,
                    inner_problem,
                    timepoints,
                    timepoints_all,
                    simulation,
                    simulation_all,
                    surrogate_all,
                    lower_bounds_all,
                    upper_bounds_all,
                    measurement_type,
                    condition_ids,
                    use_given_axes,
                )

            ax.legend()

            if not use_given_axes:
                ax.set_title(f'Group {group}, {measurement_type} data')

            ax.set_xlabel('Timepoints')
            ax.set_ylabel('Simulation/Surrogate data')

    if not use_given_axes:
        for ax in axes[len(results) :]:
            ax.remove()

    return axes


def _plot_category_rectangles_across_conditions(
    ax, category_timepoints_dict, unique_timepoints
) -> None:
    for (
        upper_bound,
        lower_bound,
    ), timepoints in category_timepoints_dict.items():
        # If the largest timepoint is not the last unique timepoint, add the next unique timepoint
        # to the list of timepoints
        max_timepoint_unique_ind = np.where(
            unique_timepoints == max(timepoints)
        )[0][0]
        if max_timepoint_unique_ind + 1 < len(unique_timepoints):
            timepoints.append(unique_timepoints[max_timepoint_unique_ind + 1])

        # Plot the category rectangle
        ax.fill_between(
            timepoints,
            [upper_bound] * len(timepoints),
            [lower_bound] * len(timepoints),
            color='gray',
            alpha=0.5,
        )

    # Add to legend meaning of gray rectangles.
    ax.fill_between(
        [],
        [],
        [],
        color='gray',
        alpha=0.5,
        label='Categories',
    )


def _plot_category_rectangles(
    ax,
    timepoints,
    upper_bounds,
    lower_bounds,
    surrogate_data,
    measurement_type,
) -> None:
    """Plot the category rectangles."""
    interval_length = 0

    for i in range(len(timepoints)):
        if i + 1 == len(timepoints) or upper_bounds[i + 1] != upper_bounds[i]:
            if i + 1 == len(timepoints):
                if upper_bounds[i] == np.inf:
                    upper_bounds[i - interval_length : i + 1] = 1.1 * max(
                        surrogate_data
                    )
                    middle_index = int((i - interval_length + i) / 2)
                    middle_timepoint = timepoints[middle_index]
                    # Draw a vertical short grey arrow at the middle point of the interval
                    # at the upper_bounds[i] height
                    ax.annotate(
                        '',
                        xy=(middle_timepoint, upper_bounds[i]),
                        xytext=(
                            middle_timepoint,
                            upper_bounds[i] + 0.1 * max(surrogate_data),
                        ),
                        arrowprops={
                            'arrowstyle': '<-',
                            'color': 'gray',
                            'linewidth': 2,
                        },
                    )
                    ax.text(
                        middle_timepoint,
                        upper_bounds[i] + 0.1 * max(surrogate_data),
                        'INF',
                        color='gray',
                        fontsize=12,
                    )
                    # Extend the ax to contain the text
                    ax.set_ylim(
                        bottom=ax.get_ylim()[0],
                        top=max(
                            ax.get_ylim()[1],
                            upper_bounds[i] + 0.1 * max(surrogate_data),
                        ),
                    )
                ax.fill_between(
                    timepoints[i - interval_length : i + 1],
                    upper_bounds[i - interval_length : i + 1],
                    lower_bounds[i - interval_length : i + 1],
                    color='gray',
                    alpha=0.5,
                )
            else:
                if upper_bounds[i] == np.inf:
                    upper_bounds[i - interval_length : i + 1] = 1.1 * max(
                        surrogate_data
                    )
                    middle_index = int((i - interval_length + i + 1) / 2)
                    middle_timepoint = timepoints[middle_index]
                    # Draw a vertical short grey arrow at the middle point of the interval
                    # at the upper_bounds[i] height
                    ax.annotate(
                        '',
                        xy=(middle_timepoint, upper_bounds[i]),
                        xytext=(
                            middle_timepoint,
                            upper_bounds[i] + 0.1 * max(surrogate_data),
                        ),
                        arrowprops={
                            'arrowstyle': '<-',
                            'color': 'gray',
                            'linewidth': 2,
                        },
                    )
                    ax.text(
                        middle_timepoint,
                        upper_bounds[i] + 0.1 * max(surrogate_data),
                        'INF',
                        color='gray',
                        fontsize=12,
                    )
                    # Extend the ax to contain the text
                    ax.set_ylim(
                        bottom=ax.get_ylim()[0],
                        top=max(
                            ax.get_ylim()[1],
                            upper_bounds[i] + 0.1 * max(surrogate_data),
                        ),
                    )

                ax.fill_between(
                    timepoints[i - interval_length : i + 2],
                    np.concatenate(
                        (
                            upper_bounds[i - interval_length : i + 1],
                            [upper_bounds[i]],
                        )
                    ),
                    np.concatenate(
                        (
                            lower_bounds[i - interval_length : i + 1],
                            [lower_bounds[i]],
                        )
                    ),
                    color='gray',
                    alpha=0.5,
                )
            interval_length = 0
        else:
            interval_length += 1
    if measurement_type == ORDINAL:
        # Add to legend meaning of rectangles
        ax.fill_between(
            [],
            [],
            [],
            color='gray',
            alpha=0.5,
            label='Categories',
        )
    elif measurement_type == CENSORED:
        # Add to legend meaning of rectangles
        ax.fill_between(
            [],
            [],
            [],
            color='gray',
            alpha=0.5,
            label='Censoring areas',
        )


def _get_data_for_plotting(
    inner_parameters: List['OptimalScalingParameter'],
    optimal_scaling_bounds: List,
    sim: List[np.ndarray],
    timepoints: List[np.ndarray],
    interval_range: float,
    interval_gap: float,
    options: Dict,
    measurement_type: str,
):
    """Return data in the form suited for plotting."""
    if options[REPARAMETERIZED] and measurement_type == ORDINAL:
        optimal_scaling_bounds = undo_inner_parameter_reparameterization(
            optimal_scaling_bounds,
            inner_parameters,
            interval_gap,
            interval_range,
        )

    simulation_all = []
    surrogate_all = []
    timepoints_all = []
    upper_bounds_all = []
    lower_bounds_all = []

    for condition_index in range(len(sim)):
        cond_simulation = []
        cond_surrogate = []
        cond_timepoints = []
        cond_upper_bounds = []
        cond_lower_bounds = []

        for inner_parameter in inner_parameters:
            if measurement_type == ORDINAL:
                upper_bound, lower_bound = get_bounds_for_category(
                    inner_parameter,
                    optimal_scaling_bounds,
                    interval_gap,
                    options,
                )
            elif measurement_type == CENSORED:
                x_category = inner_parameter.category
                lower_bound = optimal_scaling_bounds[2 * x_category - 2]
                upper_bound = optimal_scaling_bounds[2 * x_category - 1]

            # Get the condition specific simulation, mask, and timepoints
            sim_i = sim[condition_index]
            mask_i = inner_parameter.ixs[condition_index]
            t_i = timepoints[condition_index]

            y_sim = sim_i[mask_i]

            # If there is no measurement in this
            # condition for this category, skip it
            if len(y_sim) == 0:
                continue

            if mask_i.ndim == 1:
                t_sim = t_i[mask_i]
            else:
                observable_index = [
                    i for i in range(len(mask_i.T)) if any(mask_i.T[i])
                ][0]
                t_sim = timepoints[condition_index][mask_i.T[observable_index]]

            for y_sim_i in y_sim:
                if lower_bound > y_sim_i:
                    y_surrogate = lower_bound
                elif y_sim_i > upper_bound:
                    y_surrogate = upper_bound
                elif lower_bound <= y_sim_i <= upper_bound:
                    y_surrogate = y_sim_i
                else:
                    continue
                cond_surrogate.append(y_surrogate)
                cond_upper_bounds.append(upper_bound)
                cond_lower_bounds.append(lower_bound)
            cond_simulation.extend(y_sim)
            cond_timepoints.extend(t_sim)

        # Sort the surrogate datapoints and categories by timepoints, ascending.
        cond_simulation = np.array(cond_simulation)
        cond_surrogate = np.array(cond_surrogate)
        cond_timepoints = np.array(cond_timepoints)
        cond_upper_bounds = np.array(cond_upper_bounds)
        cond_lower_bounds = np.array(cond_lower_bounds)
        sort_idx = np.argsort(cond_timepoints)

        cond_simulation = cond_simulation[sort_idx]
        cond_surrogate = cond_surrogate[sort_idx]
        cond_timepoints = cond_timepoints[sort_idx]
        cond_upper_bounds = cond_upper_bounds[sort_idx]
        cond_lower_bounds = cond_lower_bounds[sort_idx]

        # Add the condition surrogate datapoints and categories to the list of all conditions.
        simulation_all.append(cond_simulation)
        surrogate_all.append(cond_surrogate)
        timepoints_all.append(cond_timepoints)
        upper_bounds_all.append(cond_upper_bounds)
        lower_bounds_all.append(cond_lower_bounds)

    return (
        simulation_all,
        surrogate_all,
        timepoints_all,
        upper_bounds_all,
        lower_bounds_all,
    )


def _get_default_axes(n_groups, **kwargs):
    """Return a list of axes with the default layout."""
    # If there is only one group, make a figure with only one plot
    if n_groups == 1:
        # Make figure with only one plot
        fig, ax = plt.subplots(1, 1, **kwargs)

        axes = [ax]
    # If there are multiple groups, make a figure with multiple plots
    else:
        # Choose number of rows and columns to be used for the subplots
        n_rows = int(np.ceil(np.sqrt(n_groups)))
        n_cols = int(np.ceil(n_groups / n_rows))

        # Make as many subplots as there are groups
        fig, axes = plt.subplots(n_rows, n_cols, **kwargs)

        # Increase the spacing between the subplots
        fig.subplots_adjust(hspace=0.35, wspace=0.25)

        # Flatten the axes array
        axes = axes.flatten()
    return axes


def _plot_observable_fit_across_conditions(
    ax,
    inner_problem,
    observable_index,
    group,
    condition_ids,
    simulation,
    simulation_all,
    surrogate_all,
    upper_bounds_all,
    lower_bounds_all,
    measurement_type,
    condition_ids_from_petab,
    use_given_axes,
):
    """Plot the observable fit across conditions.

    In case the observable has only one timepoint, the
    observable fit will be plotted against the conditions.
    """
    if measurement_type == CENSORED:
        # Get the condition indices which have censored data
        # and the corresponding condition ids with their ordering
        censored_condition_ids = [
            condition_ids[i]
            for i, cond_sim in enumerate(simulation_all)
            if len(cond_sim) > 0
        ]
        petab_censored_conditions = [
            condition_id
            for condition_id in condition_ids_from_petab
            if condition_id in censored_condition_ids
        ]
        petab_censored_conditions_ordering = [
            censored_condition_ids.index(condition_id)
            for condition_id in petab_censored_conditions
        ]
        # Get all other condition indices for quantitative data
        # and the corresponding condition ids with their ordering
        quantitative_condition_ids = [
            condition_id
            for condition_id in condition_ids
            if condition_id not in censored_condition_ids
        ]
        petab_quantitative_conditions = [
            condition_id
            for condition_id in condition_ids_from_petab
            if condition_id in quantitative_condition_ids
        ]
        petab_quantitative_condition_ordering = [
            quantitative_condition_ids.index(condition_id)
            for condition_id in petab_quantitative_conditions
        ]

    petab_condition_ordering = [
        condition_ids.index(condition_id)
        for condition_id in condition_ids_from_petab
    ]

    # Merge the simulation, surrogate, and bounds across conditions
    simulation_all = np.concatenate(simulation_all)
    surrogate_all = np.concatenate(surrogate_all)
    upper_bounds_all = np.concatenate(upper_bounds_all)
    lower_bounds_all = np.concatenate(lower_bounds_all)

    if measurement_type == CENSORED:
        # Change ordering of simulation, surrogate data and bounds to petab condition ordering
        simulation_all = simulation_all[petab_censored_conditions_ordering]
        surrogate_all = surrogate_all[petab_censored_conditions_ordering]
        upper_bounds_all = upper_bounds_all[petab_censored_conditions_ordering]
        lower_bounds_all = lower_bounds_all[petab_censored_conditions_ordering]

        whole_simulation = np.concatenate(
            [sim_i[:, observable_index] for sim_i in simulation]
        )[petab_condition_ordering]

        if not use_given_axes:
            ax.plot(
                condition_ids_from_petab,
                whole_simulation,
                linestyle='-',
                marker='.',
                color='b',
                label='Simulation',
            )
        ax.plot(
            petab_censored_conditions,
            surrogate_all,
            'rx',
            label='Surrogate data',
        )
        _plot_category_rectangles(
            ax,
            petab_censored_conditions,
            upper_bounds_all,
            lower_bounds_all,
            surrogate_all,
            measurement_type,
        )

        quantitative_data = inner_problem.groups[group][QUANTITATIVE_DATA]
        quantitative_data = quantitative_data[
            petab_quantitative_condition_ordering
        ]
        ax.plot(
            petab_quantitative_conditions,
            quantitative_data,
            'gs',
            label='Quantitative data',
        )

    elif measurement_type == ORDINAL:
        # Change ordering of simulation, surrogate data and bounds to petab condition ordering
        simulation_all = simulation_all[petab_condition_ordering]
        surrogate_all = surrogate_all[petab_condition_ordering]
        upper_bounds_all = upper_bounds_all[petab_condition_ordering]
        lower_bounds_all = lower_bounds_all[petab_condition_ordering]

        # Plot the categories and surrogate data across conditions
        if not use_given_axes:
            ax.plot(
                condition_ids_from_petab,
                simulation_all,
                linestyle='-',
                marker='.',
                color='b',
                label='Simulation',
            )
        ax.plot(
            condition_ids_from_petab,
            surrogate_all,
            'rx',
            label='Surrogate data',
        )

        _plot_category_rectangles(
            ax,
            condition_ids_from_petab,
            upper_bounds_all,
            lower_bounds_all,
            surrogate_all,
            measurement_type,
        )

    # Set the condition xticks on an angle
    ax.tick_params(axis='x', rotation=25)
    ax.legend()
    if not use_given_axes:
        ax.set_title(f'Group {group}, {measurement_type} data')

    ax.set_xlabel('Conditions')
    ax.set_ylabel('Simulation/Surrogate data')


def _plot_observable_fit_for_one_condition(
    ax,
    observable_index,
    group,
    inner_problem,
    timepoints,
    timepoints_all,
    simulation,
    simulation_all,
    surrogate_all,
    lower_bounds_all,
    upper_bounds_all,
    measurement_type,
    use_given_axes,
):
    """Plot the observable fit in case it has one condition."""
    if measurement_type == ORDINAL:
        if not use_given_axes:
            ax.plot(
                timepoints_all[0],
                simulation_all[0],
                linestyle='-',
                marker='.',
                color='b',
                label='Simulation',
            )
    elif measurement_type == CENSORED:
        quantitative_data = inner_problem.groups[group][QUANTITATIVE_DATA]
        quantitative_ixs = inner_problem.groups[group][QUANTITATIVE_IXS]
        quantitative_timepoints = timepoints[0][
            quantitative_ixs[0].T[observable_index]
        ]

        if not use_given_axes:
            ax.plot(
                timepoints[0],
                simulation[0][:, observable_index],
                linestyle='-',
                marker='.',
                color='b',
                label='Simulation',
            )
        ax.plot(
            quantitative_timepoints,
            quantitative_data,
            'gs',
            label='Quantitative data',
        )

    ax.plot(
        timepoints_all[0],
        surrogate_all[0],
        'rx',
        label='Surrogate data',
    )

    # Plot the categorie rectangles
    _plot_category_rectangles(
        ax,
        timepoints_all[0],
        upper_bounds_all[0],
        lower_bounds_all[0],
        surrogate_all[0],
        measurement_type,
    )


def _plot_observable_fit_for_multiple_conditions(
    ax,
    observable_index,
    group,
    inner_problem,
    timepoints,
    timepoints_all,
    simulation,
    simulation_all,
    surrogate_all,
    lower_bounds_all,
    upper_bounds_all,
    measurement_type,
    condition_ids,
    use_given_axes,
):
    """Plot the observable fit in case it has multiple conditions."""
    # Get the colors from the plotted simulations
    if use_given_axes:
        colors = []
        for line in ax.lines:
            if 'simulation' in line.get_label():
                colors.append(line.get_color())
    # Get as many colors as there are conditions
    else:
        colors = plt.cm.rainbow(np.linspace(0, 1, len(simulation_all)))

    if measurement_type == CENSORED:
        quantitative_data_flattened = inner_problem.groups[group][
            QUANTITATIVE_DATA
        ]
        quantitative_ixs = inner_problem.groups[group][QUANTITATIVE_IXS]
        quantitative_timepoints = [
            timepoints[cond_i][quantitative_ixs[cond_i].T[observable_index]]
            for cond_i in range(len(timepoints))
        ]
        quantitative_data = []
        index_offset = 0
        # Separate quantitative data across conditions to be as timepoints
        for cond_i in range(len(timepoints)):
            quantitative_data.append(
                quantitative_data_flattened[
                    index_offset : index_offset
                    + len(quantitative_timepoints[cond_i])
                ]
            )

    # Plot the categories and surrogate data for all conditions.
    for condition_index, condition_id, color in zip(
        range(len(simulation_all)), condition_ids, colors
    ):
        # Plot the categories and surrogate data for the current condition
        if measurement_type == ORDINAL:
            if not use_given_axes:
                ax.plot(
                    timepoints_all[condition_index],
                    simulation_all[condition_index],
                    linestyle='-',
                    marker='.',
                    color=color,
                    label=condition_id,
                )
        elif measurement_type == CENSORED:
            if not use_given_axes:
                ax.plot(
                    timepoints[condition_index],
                    simulation[condition_index][:, observable_index],
                    linestyle='-',
                    marker='.',
                    color=color,
                    label=condition_id,
                )
            ax.plot(
                quantitative_timepoints[condition_index],
                quantitative_data[condition_index],
                marker='s',
                color=color,
            )

        ax.plot(
            timepoints_all[condition_index],
            surrogate_all[condition_index],
            'x',
            color=color,
        )

    # Get all unique timepoints in ascending order
    unique_timepoints = np.unique(np.concatenate(timepoints_all))

    # Gather timepoints for each category in a dictionary
    # with upper, lower bound tuple as key and list of timepoints as value
    category_timepoints_dict = {}

    for condition_idx in range(len(simulation_all)):
        for upper_bound, lower_bound, timepoint in zip(
            upper_bounds_all[condition_idx],
            lower_bounds_all[condition_idx],
            timepoints_all[condition_idx],
        ):
            if (
                upper_bound,
                lower_bound,
            ) not in category_timepoints_dict:
                category_timepoints_dict[(upper_bound, lower_bound)] = [
                    timepoint
                ]
            else:
                category_timepoints_dict[(upper_bound, lower_bound)].append(
                    timepoint
                )

    # Plot the category rectangles
    _plot_category_rectangles_across_conditions(
        ax,
        category_timepoints_dict,
        unique_timepoints,
    )

    # Add to legend meaning of x, and -o- markers.
    ax.plot(
        [],
        [],
        'x',
        color='black',
        label='Surrogate data',
    )
    if not use_given_axes:
        ax.plot(
            [],
            [],
            linestyle='-',
            marker='.',
            color='black',
            label='Simulation',
        )
    if measurement_type == CENSORED:
        ax.plot(
            [],
            [],
            marker='s',
            color='black',
            label='Quantitative data',
        )



import warnings
from typing import Dict, List, Optional, Sequence, Union

import matplotlib.axes
import matplotlib.pyplot as plt
import numpy as np

from ..C import AMICI_SIGMAY, AMICI_Y, CURRENT_SIMULATION, DATAPOINTS, SCIPY_X
from ..problem import Problem
from ..result import Result

try:
    import amici

    from ..hierarchical.spline_approximation.calculator import (
        SplineAmiciCalculator,
    )
    from ..hierarchical.spline_approximation.problem import SplineInnerProblem
    from ..hierarchical.spline_approximation.solver import (
        SplineInnerSolver,
        get_spline_mapped_simulations,
    )
except ImportError:
    pass


def plot_splines_from_pypesto_result(
    pypesto_result: Result, start_index=0, **kwargs
):
    """Plot the inner solutions from a pypesto result.

    Parameters
    ----------
    pypesto_result:
        The pypesto result.
    start_index:
        The index of the pypesto_result.optimize_result.list to plot.
    kwargs:
        Additional arguments to pass to the plotting function.

    Returns
    -------
    fig:
        The figure.
    ax:
        The axes.
    """

    # Get the parameters from the pypesto result for the start_index.
    x_dct = dict(
        zip(
            pypesto_result.problem.objective.x_ids,
            pypesto_result.optimize_result.list[start_index]['x'],
        )
    )

    # Get the needed objects from the pypesto problem.
    edatas = pypesto_result.problem.objective.edatas
    parameter_mapping = pypesto_result.problem.objective.parameter_mapping
    amici_model = pypesto_result.problem.objective.amici_model
    amici_solver = pypesto_result.problem.objective.amici_solver
    n_threads = pypesto_result.problem.objective.n_threads
    observable_ids = amici_model.getObservableIds()

    # Fill in the parameters.
    amici.parameter_mapping.fill_in_parameters(
        edatas=edatas,
        problem_parameters=x_dct,
        scaled_parameters=True,
        parameter_mapping=parameter_mapping,
        amici_model=amici_model,
    )

    # Simulate the model with the parameters from the pypesto result.
    inner_rdatas = amici.runAmiciSimulations(
        amici_model,
        amici_solver,
        edatas,
        num_threads=min(n_threads, len(edatas)),
    )

    # If any amici simulation failed, raise warning and return None.
    if any(rdata.status != amici.AMICI_SUCCESS for rdata in inner_rdatas):
        warnings.warn(
            'Warning: Some AMICI simulations failed. Cannot plot inner solutions.'
        )
        return None

    # Get simulation and sigma.
    sim = [rdata[AMICI_Y] for rdata in inner_rdatas]
    sigma = [rdata[AMICI_SIGMAY] for rdata in inner_rdatas]

    spline_calculator = None
    for (
        calculator
    ) in pypesto_result.problem.objective.calculator.inner_calculators:
        if isinstance(calculator, SplineAmiciCalculator):
            spline_calculator = calculator
            break

    # Get the inner solver and problem.
    inner_solver = spline_calculator.inner_solver
    inner_problem = spline_calculator.inner_problem

    inner_results = inner_solver.solve(inner_problem, sim, sigma)

    return plot_splines_from_inner_result(
        inner_problem, inner_results, observable_ids, **kwargs
    )


def plot_splines_from_inner_result(
    inner_problem: 'SplineInnerProblem',
    results: List[Dict],
    observable_ids=None,
    **kwargs,
):
    """Plot the inner solutions.

    Parameters
    ----------
    inner_problem:
        The inner problem.
    results:
        The results from the inner solver.
    kwargs:
        Additional arguments to pass to the plotting function.

    Returns
    -------
    fig:
        The figure.
    ax:
        The axes.
    """

    if len(results) != len(inner_problem.groups):
        raise ValueError(
            "Number of results must be equal to number of groups of the inner subproblem."
        )

    # Get the number of groups
    n_groups = len(inner_problem.groups)

    if n_groups == 1:
        # Make figure with only one plot
        fig, ax = plt.subplots(1, 1, **kwargs)

        axs = [ax]
    else:
        # Choose number of rows and columns to be used for the subplots
        n_rows = int(np.ceil(np.sqrt(n_groups)))
        n_cols = int(np.ceil(n_groups / n_rows))

        # Make as many subplots as there are groups
        fig, axs = plt.subplots(n_rows, n_cols, **kwargs)

        # Flatten the axes array
        axs = axs.flatten()

    # for each result and group, plot the inner solution
    for result, group in zip(results, inner_problem.groups):
        group_idx = list(inner_problem.groups.keys()).index(group)

        # For each group get the inner parameters and simulation
        xs = inner_problem.get_xs_for_group(group)

        s = result[SCIPY_X]

        inner_parameters = np.array([x.value for x in xs])
        measurements = inner_problem.groups[group][DATAPOINTS]
        simulation = inner_problem.groups[group][CURRENT_SIMULATION]

        # For the simulation, get the spline bases
        delta_c, spline_bases, n = SplineInnerSolver._rescale_spline_bases(
            self=None,
            sim_all=simulation,
            N=len(inner_parameters),
            K=len(simulation),
        )
        mapped_simulations = get_spline_mapped_simulations(
            s, simulation, len(inner_parameters), delta_c, spline_bases, n
        )

        axs[group_idx].plot(
            simulation, measurements, 'bs', label='Measurements'
        )
        axs[group_idx].plot(
            spline_bases, inner_parameters, 'g.', label='Spline knots'
        )
        axs[group_idx].plot(
            spline_bases,
            inner_parameters,
            linestyle='-',
            color='g',
            label='Spline function',
        )
        axs[group_idx].plot(
            simulation, mapped_simulations, 'r^', label='Mapped simulation'
        )
        axs[group_idx].legend()
        if observable_ids is not None:
            axs[group_idx].set_title(f'{observable_ids[group-1]}')
        else:
            axs[group_idx].set_title(f'Group {group}')

        axs[group_idx].set_xlabel('Model output')
        axs[group_idx].set_ylabel('Measurements')

    for ax in axs[len(results) :]:
        ax.remove()

    return fig, axs


def _add_spline_mapped_simulations_to_model_fit(
    result: Union[Result, Sequence[Result]],
    pypesto_problem: Problem,
    start_index: int = 0,
    axes: Optional[plt.Axes] = None,
) -> Union[matplotlib.axes.Axes, None]:
    """Visualize the spline optimized model fit.

    Adds the spline-mapped simulation to the axes given by
    `pypesto.visualize.model_fit.visualize_optimized_model_fit`.
    For further details on documentation see
    :py:func:`pypesto.visualize.model_fit.visualize_optimized_model_fit`.
    """

    # If no visualize_optimized_model_fit axes were given,
    # return None.
    if axes is None:
        return None

    # Get the parameters from the pypesto result for the start_index.
    x_dct = dict(
        zip(
            pypesto_problem.objective.x_ids,
            result.optimize_result.list[start_index]['x'],
        )
    )

    # Get the needed objects from the pypesto problem.
    edatas = pypesto_problem.objective.edatas
    parameter_mapping = pypesto_problem.objective.parameter_mapping
    amici_model = pypesto_problem.objective.amici_model
    amici_solver = pypesto_problem.objective.amici_solver
    n_threads = pypesto_problem.objective.n_threads

    # Fill in the parameters.
    amici.parameter_mapping.fill_in_parameters(
        edatas=edatas,
        problem_parameters=x_dct,
        scaled_parameters=True,
        parameter_mapping=parameter_mapping,
        amici_model=amici_model,
    )

    # Simulate the model with the parameters from the pypesto result.
    inner_rdatas = amici.runAmiciSimulations(
        amici_model,
        amici_solver,
        edatas,
        num_threads=min(n_threads, len(edatas)),
    )

    # If any amici simulation failed, raise warning and return None.
    if any(rdata.status != amici.AMICI_SUCCESS for rdata in inner_rdatas):
        warnings.warn(
            'Warning: Some AMICI simulations failed. Cannot plot inner solutions.'
        )
        return None

    # Get simulation and sigma.
    sim = [rdata[AMICI_Y] for rdata in inner_rdatas]
    sigma = [rdata[AMICI_SIGMAY] for rdata in inner_rdatas]

    spline_calculator = None
    for calculator in pypesto_problem.objective.calculator.inner_calculators:
        if isinstance(calculator, SplineAmiciCalculator):
            spline_calculator = calculator
            break

    # Get the inner solver and problem.
    inner_solver = spline_calculator.inner_solver
    inner_problem = spline_calculator.inner_problem

    # Solve the inner problem.
    inner_results = inner_solver.solve(inner_problem, sim, sigma)

    # Get the observable ids.
    observable_ids = amici_model.getObservableIds()

    for inner_result, group in zip(inner_results, inner_problem.groups):
        observable_id = observable_ids[group - 1]
        # Get the ax for the current observable.
        ax = [
            ax
            for ax in axes.values()
            if observable_id in ax.legend().get_texts()[0].get_text().split()
        ][0]

        # Get the inner parameters and simulation.
        xs = inner_problem.get_xs_for_group(group)
        s = inner_result[SCIPY_X]

        inner_parameters = np.array([x.value for x in xs])
        simulation = inner_problem.groups[group][CURRENT_SIMULATION]

        # For the simulation, get the spline bases
        delta_c, spline_bases, n = SplineInnerSolver._rescale_spline_bases(
            self=None,
            sim_all=simulation,
            N=len(inner_parameters),
            K=len(simulation),
        )
        # and the spline-mapped simulations.
        mapped_simulations = get_spline_mapped_simulations(
            s, simulation, len(inner_parameters), delta_c, spline_bases, n
        )

        # Plot the spline-mapped simulations to the ax with same color
        # and timepoints as the lines which have 'simulation' in their label.
        plotted_index = 0
        for line in ax.lines:
            if 'simulation' in line.get_label():
                color = line.get_color()
                timepoints = line.get_xdata()
                condition_mapped_simulations = mapped_simulations[
                    plotted_index : len(timepoints) + plotted_index
                ]
                plotted_index += len(timepoints)

                ax.plot(
                    timepoints,
                    condition_mapped_simulations,
                    color=color,
                    linestyle='dotted',
                    marker='^',
                )

        # Add linestyle='dotted' and marker='^' to the legend as black spline mapped simulations.
        ax.plot(
            [],
            [],
            color='black',
            linestyle='dotted',
            marker='^',
            label='Spline mapped simulation',
        )

    # Reset the legend.
    for ax in axes.values():
        ax.legend()

    return axes
